﻿#include "TCP_client.h"


/*********************************************************************************************
* 功能    ：	socket版本
* 描述	  ：	启动socket服务
* 输入    ：    无
* 输出    ：    true  启动成功
*				false 启动失败
*********************************************************************************************/
bool InitSocket_Version(void)
{
	WORD sockVersion = MAKEWORD(2, 2);//使用winsocket2.2版本
	WSADATA wsaData;
	if (WSAStartup(sockVersion, &wsaData) != 0)
	{
		return false;
	}
	return true;
}

/*********************************************************************************************
* 功能    ：	IP地址有效性检测
* 描述	  ：	检测输入的IP地址是否合法
* 输入    ：    IP   输入的IP地址
* 输出    ：    true  IP地址合法
*				false IP地址非法
*********************************************************************************************/
bool Check_IP(string IP)
{
	int s[4];
	string ip = IP;
	if (ip.length() < 7 || ip.length() > 15) //长度判定
		return false;
	if (sscanf_s(IP.c_str(), "%d.%d.%d.%d", &s[0], &s[1], &s[2], &s[3]) != 4) //IPV4格式正确
	{
		return false;
	}
	string newip = to_string(s[0]) + "." + to_string(s[1]) + "." + to_string(s[2]) + "." + to_string(s[3]);
	if (ip != newip) //前导0
		return false;
	if ((s[0] & 0xffffff00) || (s[1] & 0xffffff00) || (s[2] & 0xffffff00) || (s[3] & 0xffffff00)) //判断每一段大小是否符合要求
	{
		return false;
	}
	return true;
}

/*********************************************************************************************
* 功能    ：	获取从站IP地址和端口号
* 描述	  ：	终端输入从站IP地址和端口号
* 输入    ：    IP 地址   （IPV4）
*				*Port_number 端口号（1-65535）
* 输出    ：    无
*********************************************************************************************/
void Input_IP(string& IP, unsigned int *Port_number)
{
	int i = 1;
	do
	{
		if (i == 1)
		{
			cout << "请输入从站IP：";
			i = 0;
		}
		else
			cout << "IP地址格式不正确，请重新输入从站IP：";
		cin >> IP;
	} while (!Check_IP(IP));

	i = 1;
	cout << "IP地址输格式入正确，请输入从站端口号：";
	do
	{
		if (i != 1)
			cout << "请重新输入从站端口号：";
		cin >> *Port_number;
		i = 0;
	} while (*Port_number == 0 || *Port_number > 65535); //端口不能为0 ，端口号范围1---65535

}

/*********************************************************************************************
* 功能    ：	初始化客户端
* 描述	  ：	根据终端输入从站IP地址和端口号连接对应的服务器,设置超时时间
* 输入    ：    IP 地址   （IPV4）
*				Port_number 端口号（1-65535）
* 输出    ：    ClientSocket 连接成功的套接字
*********************************************************************************************/
SOCKET Init_client()
{
	string IP;
	unsigned int Port_number;
	Input_IP(IP, &Port_number);
	if (InitSocket_Version() == 0)
		return INVALID_SOCKET;
	SOCKET ClientSocket = socket(AF_INET, SOCK_STREAM, 0);
	sockaddr_in Serversock_in;
	Serversock_in.sin_addr.S_un.S_addr = inet_addr(IP.c_str());
	Serversock_in.sin_family = AF_INET;
	Serversock_in.sin_port = htons(Port_number);
	if(SOCKET_ERROR == connect(ClientSocket, (SOCKADDR*)&Serversock_in, sizeof(SOCKADDR)))
	{
		cout << "尝试连接TCP从站失败" << endl;
		return INVALID_SOCKET;
	}
	cout << "连接TCP从站成功" << endl;
	TIMEVAL timeout;
	timeout.tv_sec = 60000;  //ms
	timeout.tv_usec = 0;  //us
	setsockopt(ClientSocket, SOL_SOCKET, SO_RCVTIMEO, (char *)&timeout, sizeof(struct timeval));//设置接收超时时间
	return ClientSocket;
}

/*********************************************************************************************
* 功能    ：	打印读取从站的线圈状态
* 描述	  ：	对响应报文中的数据进行解析和显示
* 输入    ：    Response_Message 响应报文  Request_Message 请求报文
* 输出    ：    无
*********************************************************************************************/
void Printf_Coil_date(UINT8 *Response_Message, UINT8 *Request_Message)
{
	printf("从站设备ID %02X 功能码为 %02X\n", Response_Message[6], Response_Message[7]);
	unsigned int temp1 = ((Request_Message[8] << 8) | Request_Message[9]);
	unsigned int Number = ((Request_Message[10] << 8) | Request_Message[11]);
	printf("线圈起始地址为%d \n",temp1);
	unsigned int temp = temp1;
	for (int i = 0; i < Response_Message[8]; i++)
	{
		unsigned int temp2 = temp + 7;
		if (temp2 > temp1 + Number - 1)
			temp2 = temp1 + Number - 1;
		printf("线圈%d --- %d的状态为：%02X \n", temp2, temp, Response_Message[9+i]);
		temp = temp + 8;
	}
}

/*********************************************************************************************
* 功能    ：	打印读取从站的寄存器状态
* 描述	  ：	对响应报文中的数据进行解析和显示
* 输入    ：    Response_Message 响应报文  Request_Message 请求报文
* 输出    ：    无
*********************************************************************************************/
void Printf_Register_date(UINT8 *Response_Message, UINT8 *Request_Message)
{
	printf("从站设备ID %02X 功能码为 %02X\n", Response_Message[6], Response_Message[7]);
	unsigned int temp1 = ((Request_Message[8] << 8) | Request_Message[9]);
	unsigned int Number = ((Request_Message[10] << 8) | Request_Message[11]);
	printf("寄存器起始地址为%d \n", temp1);
	for (int i = 0; i < Response_Message[8]; i = i + 2)
	{
		printf("寄存器%d的值为：%02X %02X \n", temp1++, Response_Message[9 + i], Response_Message[10+i]);
	}
}

/*********************************************************************************************
* 功能    ：	打印读取异常响应报文信息
* 描述	  ：	对响应报文中的数据进行解析和显示
* 输入    ：    Response_Message 响应报文
* 输出    ：    无
*********************************************************************************************/
void Printf_Anomaly_date(UINT8 *Response_Message)
{
	UINT8 a = Response_Message[8];
	printf("从站设备ID %02X 功能码为 %02X\n", Response_Message[6], Response_Message[7]);
	switch (a)
	{
	case 0x01:	printf("%02X : 从站设备不支持此功能码\n",a); break;
	case 0x02:	printf("%02X : 指定的数据地址在从站设备中不存在\n",a); break;
	case 0x03:	printf("%02X : 指定的数据超过范围或者不允许使用\n",a); break;
	case 0x04:	printf("%02X : 从站设备处理响应的过程中，出现未知错误等\n",a); break;
	default:	printf("Unkown Other Error!!!!!\n");
	}
}



/*********************************************************************************************
* 功能    ：	检测响应报文长度
* 描述	  ：	对响应报文中存放长度字节进行判断和请求报文对比
* 输入    ：    Response_Message 响应报文 Request_Message 请求报文 Response_Message_len 接收到的数据长度
* 输出    ：    true 长度正常  false 长度异常
*********************************************************************************************/
bool Check_Response_Message_len(UINT8 *Response_Message, UINT8 *Request_Message, int Response_Message_len)
{
	if ((Response_Message_len - 6) != Response_Message[5])
		return false;
	unsigned int Operations_Number = Request_Message[10] << 8 | Request_Message[11];
	unsigned int Respone_Len = Count_Respone_Len(Request_Message[7], Operations_Number);
	if (Request_Message[7] == 0x0F || Request_Message[7] == 0x10)
		if (Response_Message_len != 12)
			return false;
	if (Request_Message[7] == 0x01 || Request_Message[7] == 0x03)
	{
		if (Respone_Len + 9 != Response_Message_len)
			return false;
		if (Respone_Len != Response_Message[8])
			return false;
	}
	return true;
}
/*********************************************************************************************
* 功能    ：	判断响应报文是否可以正常解析
* 描述	  ：	通过长度和异常码等判定该响应报文是否可以解析
* 输入    ：    Response_Message 响应报文 Request_Message请求报文  Response_Message_len 响应报文长度
* 输出    ：    true可以正常解析 false 无法解析
*********************************************************************************************/
bool Analysis_Response_Message(UINT8 *Response_Message, UINT8 *Request_Message, int Response_Message_len)
{
	if (Response_Message_len < 9)
		return false;
	for (int i = 0; i < 7; i++)
	{
		if (i == 4 || i == 5) //后续字节长度
			continue;
		if (Response_Message[i] != Request_Message[i]) //0 1 传输标识，2 3 协议标识，6设备ID 
			return false;  //一旦不一致，判定为异常报文数据
	}
	if (Response_Message[4] != 0x00) //4 固定0x00
		return false;

	if (Response_Message[7] == Request_Message[7] + 0x80)//先处理异常响应
	{
		if (Response_Message_len == 9)
		{
			Printf_Anomaly_date(Response_Message);
			return true;
		}
		else
			return false;
	}
	if (Response_Message[7] != Request_Message[7]) //功能码判断
		return false;
	if (Request_Message[7] == 0x0F || Request_Message[7] == 0x10) //0F 10 地址和操作数量判定
	{
		for (int j = 8; j < 12; j++)
		{
			if (Request_Message[j] != Response_Message[j])
				return false;
		}
	}
	if (!Check_Response_Message_len(Response_Message, Request_Message, Response_Message_len))
		return false;
	if (Response_Message[7] == 0x01)
		Printf_Coil_date(Response_Message, Request_Message);
	else if (Response_Message[7] == 0x03)
		Printf_Register_date(Response_Message, Request_Message);
	else if (Response_Message[7] == 0x0F)
		printf("成功写入从站线圈%d个\n", ((Request_Message[10] << 8) | Request_Message[11]));
	else if (Response_Message[7] == 0x10)
		printf("成功写入从站寄存器%d个\n", ((Request_Message[10] << 8) | Request_Message[11]));
	else
		return false;
	return true;
}

void Clear_buf(SOCKET ClientSocket)
{
	unsigned long bytesToRecv;
	char temp[500];
	do
	{

		ioctlsocket(ClientSocket, FIONREAD, &bytesToRecv);
		if (bytesToRecv != 0)//不等于0时进行清理操作
		{
			if (bytesToRecv > 500)
			{
				recv(ClientSocket, temp, 500, 0);
			}
			else
				recv(ClientSocket, temp, bytesToRecv, 0);
		}
	} while (bytesToRecv != 0);

}

/*********************************************************************************************
* 功能    ：	运行客户端
* 描述	  ：	根据终端输入从站IP地址和端口号连接对应的服务器并且生成发送请求等待响应
* 输入    ：    无
* 输出    ：    无
*********************************************************************************************/
int Tcp_client()
{
	SOCKET ClientSocket;
	do
	{
		ClientSocket = Init_client();	
	} while (INVALID_SOCKET == ClientSocket);
	UINT8 Request_Message[600];
	UINT8 Response_Message[600];
	TIMEVAL timeout;
	timeout.tv_sec = 0;
	timeout.tv_usec = 0;
	fd_set  reads;
	UINT16 Transmission_Indicator = 0x00;
	while (true)
	{
		int Function_code = Input_Function_code();
		unsigned int Operations_Number = Input_Operations_Number(Function_code);
		unsigned int Starting_address = Input_Starting_address();
		string Write_date = Input_Write_date(Function_code, Operations_Number);
		int Request_Message_len = Crate_TCP_Message(Request_Message, Function_code, Operations_Number, Starting_address, Write_date);
		Request_Message[0] = Transmission_Indicator >> 8;
		Request_Message[1] = (UINT8)Transmission_Indicator;
		printf("主站请求 ：");
		for (int i = 0; i < Request_Message_len; i++)
		{
			printf("%02x  ", Request_Message[i]);
		}
		printf("\n");
		Log_Note(Request_Message, 1, Request_Message_len);
		Clear_buf(ClientSocket); //清理缓冲区
		int test = send(ClientSocket, (char*)Request_Message, Request_Message_len, 0);
		if (Transmission_Indicator == 0xffff)
		{
			Transmission_Indicator = 0x00;
		}
		Transmission_Indicator++;
		if (test == (-1))
		{
			printf("连接异常，请检查连接状态。\n");
			printf("**************************** Press Enter To Contioun ****************************\n");
			getchar();
			system("cls");
			closesocket(ClientSocket);
			WSACleanup();
			printf("是否重新连接服务器： 1 重新连接  0 关闭本软件");
			int a = 0;
			do
			{
				cin >> a;
				cin.clear();
				cin.sync();
			} while (!(a == 0 || a == 1));
			if (a == 0)
				break;
			system("cls");
			do
			{
				ClientSocket = Init_client();
			} while (INVALID_SOCKET == ClientSocket);
			system("cls");
			continue;
		}
		memset(Response_Message, 0, 600);
		FD_ZERO(&reads);
		FD_SET(ClientSocket, &reads);
		int Response_Message_len = recv(ClientSocket, (char*)Response_Message, 600, 0);
		if (Response_Message_len > 0)
		{
			Log_Note(Response_Message, 0, Response_Message_len);
			printf("从站响应 ：");
			for (int i = 0; i < Response_Message_len; i++)
			{
				printf("%02x  ", Response_Message[i]);
			}
			printf("\n");
			if (!Analysis_Response_Message(Response_Message, Request_Message, Response_Message_len))
				printf("响应报文异常\n\n");
		}
		else if (select(0, &reads, 0, 0, &timeout))
		{
			printf("连接异常，请检查连接状态。\n");
			printf("**************************** Press Enter To Contioun ****************************\n");
			getchar();
			system("cls");
			closesocket(ClientSocket);
			WSACleanup();
			printf("是否重新连接服务器： 1 重新连接  0 关闭本软件\n");
			int a = 0;
			do
			{
				cin >> a;
				cin.clear();
				cin.sync();
			} while (!(a == 0 || a == 1));
			if (a == 0)
				break;
			system("cls");
			do
			{
				ClientSocket = Init_client();
			} while (INVALID_SOCKET == ClientSocket);
			system("cls");
			continue;
		}
		else
			cout << "响应超时" << endl;
		printf("**************************** Press Enter To Contioun ****************************\n");
		getchar();
		system("cls");
	}
	//关闭套接字
	closesocket(ClientSocket);
	//关闭服务
	WSACleanup();
	return 0;
}
