﻿#include "common.h"


/*********************************************************************************************
* 功能    ：	选择功能码
* 描述	  ：	在功能码 0x01 0x03 0x0F 0x10中选择一个功能码
* 输入    ：    无
* 输出    ：    Function_code  选择的功能码
*********************************************************************************************/
int Input_Function_code(void)
{
	int Function_code;
	cin >> hex;
	int i = 1;
	printf("功能码 ：  0x01      0x03        0x0F         0x10 \n");
	printf("说明   ： 读线圈   读寄存器   写多个线圈  写多个寄存器\n");
	do
	{
		if (i == 1)
		{
			i = 0;
			cout << "请输入选择的功能码：";
		}
		else
			cout << "输入功能码错误，请重新选择的功能码";
		cin >> Function_code;
		cin.clear();
		cin.sync();
	} while (Function_code != 0x01 && Function_code != 0x03 && Function_code != 0x0F && Function_code != 0x10);
	return Function_code;
}

/*********************************************************************************************
* 功能    ：	输入操作起始地址码
* 描述	  ：	起始地址码，最大值65535
* 输入    ：    无
* 输出    ：    Starting_address 操作起始地址码
*********************************************************************************************/
unsigned int Input_Starting_address(void)
{
	unsigned int Starting_address;
	int i = 1;
	cin >> dec;
	do
	{
		if (i == 1)
		{
			i = 0;
			cout << "请输入操作(十进制)起始地址：";
		}
		else
			cout << "请重新输入操作(十进制)起始地址";
		cin >> Starting_address;
		cin.clear();
		cin.sync();
	} while (Starting_address > 65535 || Starting_address < 0);
	return Starting_address;
}

/*********************************************************************************************
* 功能    ：	输入操作数量
* 描述	  ：	操作数量，0x01 1-200  0x03 1-125  0x0xF 1-1968 0x10 1-123
* 输入    ：    Function_code 选择的功能码类型
* 输出    ：    Operations_Number 对应功能码类型的操作数量
*********************************************************************************************/
unsigned int Input_Operations_Number(int Function_code)
{
	unsigned int Operations_Number;
	int i = 1;
	bool flage = true;
	printf("  功能码  0x01    0x03    0x0F   0x10\n");
	printf("操作数量  1-2000  1-125  1-1968  1-123\n");
	cin >> dec;
	do
	{
		if (i == 1)
		{
			i = 0;
			cout << "请输入操作(十进制)数量：";
		}
		else
			cout << "请重新输入操作(十进制)数量";
		cin >> Operations_Number;
		cin.clear();
		cin.sync();
		if (Function_code == 0x01 && Operations_Number >= 1 && Operations_Number <= 65535)
			flage = false;
		if (Function_code == 0x03 && Operations_Number >= 1 && Operations_Number <= 65535)
			flage = false;
		if (Function_code == 0x0F && Operations_Number >= 1 && Operations_Number <= 65535)
			flage = false;
		if (Function_code == 0x10 && Operations_Number >= 1 && Operations_Number <= 65535)
			flage = false;
	} while (flage);
	return Operations_Number;
}

/*********************************************************************************************
* 功能    ：	检测写入数据
* 描述	  ：	检测写入数据格式是否正确
* 输入    ：    Write_date 写入的数据 Write_date_number写入数据的组数
* 输出    ：    true  写入数据格式正确
*				false 写入数据格式错误
*********************************************************************************************/
bool Check_Write_date(string Write_date, int Write_date_number)
{
	int Write_date_len = 2 * Write_date_number + Write_date_number - 1;
	if (Write_date.length() != Write_date_len)
		return false;
	int Space_number = 0;
	int j = -1;
	for (unsigned int i = 0; i < Write_date.length(); i++)
	{
		if (Write_date[i] == 0x20)//空格位置判定
		{
			if (i - j == 3)
			{
				Space_number++;
				j = i;
				continue;
			}
			else
				return false;
		}
		if ((Write_date[i] >= 0x30 && Write_date[i] <= 0x39)) // 1-9
			continue;
		else if ((Write_date[i] >= 0x41 && Write_date[i] <= 0x5A)) //A-Z
			continue;
		else if ((Write_date[i] >= 0x61 && Write_date[i] <= 0x7A)) //a-z
			continue;
		else
			return false;
	}
	if (Space_number != Write_date_number - 1)
		return false;
	return true;
}


/*********************************************************************************************
* 功能    ：	计算写入数据的字节数
* 描述	  ：	通过对应的功能码和操作数量计算对应的数据字节数
* 输入    ：    Function_code 选择的功能码类型 Operations_Number对应功能码类型的操作数量
* 输出    ：    Write_date_number 写入数据的字节数
*********************************************************************************************/
unsigned int Count_Write_date_number(int Function_code, unsigned int Operations_Number)
{
	unsigned int Write_date_number = 0;
	if (Function_code == 0x01 || Function_code == 0x03)
		return Write_date_number;
	if (Function_code == 0x0F) //读线圈
	{
		Write_date_number = Operations_Number / 8;
		if (Operations_Number % 8)
			Write_date_number++;
	}
	if (Function_code == 0x10)
		Write_date_number = Operations_Number * 2;
	return Write_date_number;
}

/*********************************************************************************************
* 功能    ：	计算请求报文计算预期响应报文字节数
* 描述	  ：	通过对应的功能码和操作数量计算对应的数据字节数
* 输入    ：    Function_code 选择的功能码类型 Operations_Number对应功能码类型的操作数量
* 输出    ：    Respone_len 预期响应的数据字节数
*********************************************************************************************/
unsigned int Count_Respone_Len(int Function_code, unsigned int Operations_Number)
{
	unsigned int Respone_len = 0;
	if (Function_code == 0x0F|| Function_code == 0x10)
		return Respone_len;
	if (Function_code == 0x01) //读线圈
	{
		Respone_len = Operations_Number / 8;
		if (Operations_Number % 8)
			Respone_len++;
	}
	if (Function_code == 0x03)
		Respone_len = Operations_Number * 2;
	return Respone_len;
}

/*********************************************************************************************
* 功能    ：	输入写入数据
* 描述	  ：	根据写入数量和功能码类型输入相应的数据
* 输入    ：    Function_code 选择的功能码类型 Operations_Number对应功能码类型的操作数量
* 输出    ：    Write_date 写入的数据 
*********************************************************************************************/
string Input_Write_date(int Function_code, unsigned int Operations_Number)
{
	int i = 1;
	string Write_date;
	unsigned int Write_date_number = Count_Write_date_number(Function_code, Operations_Number);
	if (Function_code == 0x01 || Function_code == 0x03)
		return Write_date;//读线圈/寄存器不需要输入写入数据
	
	printf("请输入N组写入数据（2位16进制为一组，空格间隔）例如：00 00 01 00 20 05\n");
	do
	{
		if (i == 1)
		{
			i = 0;
			printf("请输入%d组写入数据 ：", Write_date_number);
		}
		else
			printf("请重新输入%d组写入数据 ：", Write_date_number);
		getline(cin, Write_date);
		cin.clear();
		cin.sync();
	} while (!Check_Write_date(Write_date, Write_date_number));

	return Write_date;
}

/*********************************************************************************************
* 功能    ：	生成MBAP报头
* 描述    ： MBAP报文头的包括的内容：                                                   
* +-------------+---------+--------+--------+------------------------------+ 
* |     域      |  长度   | 客户机 | 服务器 |              描述            | 
* +-------------+---------+--------+--------+------------------------------+ 
* |事务元标识符 | 2个字节 |  启动  |  复制  |请求/响应事务处理的识别码     | 
* +-------------+---------+--------+--------+------------------------------+ 
* |协议标识符   | 2个字节 |  启动  |  复制  |0=MODBUS 协议                 | 
* +-------------+---------+--------+--------+------------------------------+ 
* |    长度     | 2个字节 |  启动  |  启动  |以下字节的数量                | 
* +-------------+---------+--------+--------+------------------------------+ 
* |单元标识符   | 1个字节 |  启动  |  复制  |连接的远程从站的识别码        | 
* 输入    ：    无
* 输出    ：    MBAP报文头内容
*********************************************************************************************/
void Crate_MBAP(UINT8 *Message, int Function_code, unsigned int Operations_Number)
{
	Message[0] = 0x00;
	Message[1] = 0x00;
	Message[2] = 0x00;
	Message[3] = 0x00;
	Message[4] = 0x00;
	Message[5] = Count_Write_date_number(Function_code, Operations_Number) + 0x06;
	if (Function_code == 0x0F || Function_code == 0x10)
	{
		Message[5] = Message[5] + 1;
	}
	Message[6] = DEVICE_ID;
}

/*********************************************************************************************
* 功能    ：	字符串转UINT8类型
* 描述	  ：	根据写入数量和功能码类型输入相应的数据
* 输入    ：    *Message 消息帧存放的数组   Write_date 要转换的字符串   
*				Message_len消息帧数组的起始位置
* 输出    ：    Message_len 消息帧的长度
*********************************************************************************************/
int HexStringtoByte(UINT8 *Message, string Write_date, int Message_len, int Function_code, unsigned int Operations_Number)
{
	if (Write_date.length() == 0)
		return Message_len;
	Message[Message_len] = Count_Write_date_number(Function_code, Operations_Number);
	Message_len++;
	const char *b = Write_date.c_str();
	for (unsigned int i = 0; i + 3 < Write_date.length(); i = i + 3)
	{
		sscanf_s(b + i, "%02X", (int *)(Message + Message_len));
		Message_len++;
	}
	sscanf_s(b + Write_date.length() - 2, "%02X", (int *)(Message + Message_len));
	return ++Message_len;
}

/*********************************************************************************************
* 功能    ：	生成TCP模式下的消息帧
* 描述	  ：	根据要写入的数据生成消息帧
* 输入    ：    *Message 消息帧存放的数组   Write_date 写入数据的字符串
*				Function_code 功能码   Operations_Number 操作数量  Starting_address起始地址
* 输出    ：    消息帧的总长度
*********************************************************************************************/
int Crate_TCP_Message(UINT8 *Message, int Function_code, unsigned int Operations_Number, unsigned int Starting_address, string Write_date)
{

	Crate_MBAP(Message, Function_code, Operations_Number);
	Message[7] = Function_code;//功能码
	Message[8] = Starting_address >> 8;//起始地址H
	Message[9] = Starting_address;//起始地址L
	Message[10] = Operations_Number >> 8;//操作数量H
	Message[11] = Operations_Number;//操作数量L

	return HexStringtoByte(Message, Write_date, 12, Function_code, Operations_Number);
	
}

/*********************************************************************************************
* 功能    ：	日志记录
* 描述	  ：	记录每次通信的请求和响应报文
* 输入    ：    *Message 消息帧存放的数组   flage = 1 请求报文  flage = 其他 响应报文
*				Message_len 报文数据长度
* 输出    ：    无
*********************************************************************************************/

void Log_Note(UINT8 *Message, int flage, int Message_len)
{
	FILE *fp = NULL;
	time_t timep;
	struct tm *p;
	time(&timep);
	p = gmtime(&timep);
	string Recv_str = to_string(1900 + p->tm_year) + "-" + to_string(1 + p->tm_mon) + "-" + to_string(p->tm_mday)
		+ " " + to_string(8 + p->tm_hour) + ":" + to_string(p->tm_min) + ":" + to_string(p->tm_sec) + " Recv:";
	string Send_str= to_string(1900 + p->tm_year) + "-" + to_string(1 + p->tm_mon) + "-" + to_string(p->tm_mday)
		+ " " + to_string(8 + p->tm_hour) + ":" + to_string(p->tm_min) + ":" + to_string(p->tm_sec) + " Send:";

	/* 打开文件用于读写 */
	if ((fp = fopen("Modbus_TCP_log.txt", "a+")) == NULL)
	{
		printf("打开文件失败");
		return ;
	}

	if (flage == 1)
	{
		fwrite(Send_str.c_str(), Send_str.length(), 1, fp);
		for (int i = 0; i < Message_len; i++)
		{
			fprintf(fp, "%02X ", Message[i]);
		}
		fprintf(fp,"\n");
	}
	else
	{
		fwrite(Recv_str.c_str(), Recv_str.length(), 1, fp);
		for (int i = 0; i < Message_len; i++)
		{
			fprintf(fp, "%02X ", Message[i]);
		}
		fprintf(fp, "\n");
	}
	fclose(fp);
	fp = NULL;
}
